/* gtkentryextra.c - GtkEntryExtra widget implementation
 *
 * Authors:         Tim Janik
 *                  Kristian Rietveld <kristian@planet.nl>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "gtkentryextra.h"

#include <gdk/gdkkeysyms.h>
#include <gtk/gtkbindings.h>
#include <gtk/gtkobject.h>
#include <gtk/gtksignal.h>

#include <string.h>

enum {
  AUTOCOMPLETED,
  LAST_SIGNAL
};

static void gtk_entry_extra_class_init        (GtkEntryExtraClass *class);
static void gtk_entry_extra_init              (GtkEntryExtra      *entrye);

static gboolean  gtk_entry_extra_match        (GtkEntryExtra      *entrye,
					       const gchar        *string);
static gchar    *gtk_entry_extra_try_complete (GtkEntryExtra      *entrye);
static void      gtk_entry_extra_add_history  (GtkEntryExtra      *entrye,
					       const gchar        *string);
static gchar    *gtk_entry_extra_intersect    (gint                max_len,
					       GSList             *strings);
static gint      gtk_entry_extra_key_press    (GtkWidget          *widget,
					       GdkEventKey        *event);


static gint gtk_entry_extra_signals[LAST_SIGNAL] = {0};
static GtkEntryExtra *gtk_entry_extra_parent_class = NULL;

/* functions */

static void
gtk_entry_extra_class_init (GtkEntryExtraClass *class)
{
  g_return_if_fail (class != NULL);
  
  gtk_entry_extra_parent_class = g_type_class_peek_parent (class);

  gtk_entry_extra_signals[AUTOCOMPLETED] =
    gtk_signal_new ("autocompleted",
		    GTK_RUN_LAST,
		    gtk_entry_extra_get_type (),
		    GTK_SIGNAL_OFFSET (GtkEntryExtraClass, autocompleted),
		    gtk_signal_default_marshaller,
		    GTK_TYPE_NONE, 0);
}

static void
gtk_entry_extra_init (GtkEntryExtra *entrye)
{
  g_return_if_fail (entrye != NULL);
  g_return_if_fail (GTK_IS_ENTRY_EXTRA (entrye));

  entrye->cboxtext = NULL;
  
  entrye->autocomplete = TRUE;
  entrye->pattern_matching = entrye->keep_history = FALSE;
  entrye->history_length = 10;
}

static gboolean
gtk_entry_extra_match (GtkEntryExtra *entrye,
		       const gchar   *string)
{
  int      i = 0;
  int      smallest = 0;
  int      quitloop = 0;
  gchar   *match = NULL;
  GSList  *s = NULL, *strings = NULL, *matches = NULL;
  gchar   *tmp = (gchar *)g_malloc0 (sizeof (gchar) * (strlen (string) + 1));
  
  if (strlen (string) < 1)
    return FALSE;
  
  strings = (GSList *)gtk_combo_box_text_build_item_list
    (GTK_TREE_MODEL (GTK_COMBO_BOX_TEXT (entrye->cboxtext)->model));
  
  if (!strings)
    return FALSE;
  
  for (s = strings; s != NULL; s = s->next)
    {
      gchar *tmpdata = NULL;

      if (!s->data)
	continue;

      tmpdata = (gchar *)s->data;

      strncpy (tmp, tmpdata, strlen(string));

      if (!strcmp (tmp, string))
        {
	  match = g_strdup (s->data);
	  matches = g_slist_append (matches, match);
        }
    }
  
  g_slist_free (strings);
  g_free (tmp);
  
  if (!matches)
    return FALSE;
  
  if (g_slist_length (matches) == 1 && matches->data != NULL)
    {
      gtk_entry_set_text (GTK_ENTRY (entrye), (gchar *)matches->data);
      gtk_entry_set_position (GTK_ENTRY (entrye),
			      GTK_ENTRY (entrye)->text_length);
      
      g_slist_free (matches);
      
      return TRUE;
    }

  smallest = strlen ((gchar *)matches->data);
  s = matches->next;

  for (; s != NULL; s = s->next)
    if (strlen ((gchar *)s->data) < smallest)
      smallest = strlen ((gchar *)s->data);

  while (!quitloop)
    {
      gchar   curchar;

      if (i >= smallest) {
	quitloop = 1;
	break;
      }

      curchar = ((gchar *)matches->data)[i];
      
      for (s = matches; s != NULL; s = s->next)
	{
	  if (((gchar *)s->data)[i] != curchar)
	    {
	      quitloop = 1;
	      break;
	    }
	}
      
      i++;
    }
  
  if (i > 0 && i > strlen (string))
    {
      gchar *tmpstr;
      
      tmpstr = g_malloc0 (sizeof (gchar) * (i));
      strncpy (tmpstr, ((gchar *)matches->data), i - 1);
      tmpstr[i] = 0;
      
      gtk_entry_set_text (GTK_ENTRY (entrye), tmpstr);
      gtk_entry_set_position (GTK_ENTRY (entrye),
			      GTK_ENTRY (entrye)->text_length);
      
      g_free (tmpstr);
    }
  
  g_slist_free (matches);
  
  return FALSE;
}

static gchar *
gtk_entry_extra_try_complete (GtkEntryExtra *entrye)
{
  gint     max_len = 0;
  gchar   *completion;
  GSList  *s = NULL;
  GSList  *strings = NULL;

  g_return_val_if_fail (entrye != NULL, NULL);
  g_return_val_if_fail (GTK_IS_ENTRY_EXTRA (entrye), NULL);
  g_return_val_if_fail (entrye->cboxtext != NULL, NULL);
  
  strings = gtk_combo_box_text_build_item_list 
    (GTK_TREE_MODEL(GTK_COMBO_BOX_TEXT (entrye->cboxtext)->model));
  
  if (!strings)
    return NULL;
  
  for (s = strings; s != NULL; s = s->next)
    if (s->data)
      max_len = MAX (max_len, strlen ((gchar *)s->data));
  
  completion = gtk_entry_extra_intersect (max_len, strings);
  g_slist_free (strings);

  return completion;
}

static void
gtk_entry_extra_add_history (GtkEntryExtra *entrye,
			     const gchar *string)
{
  gint childs;
  
  g_return_if_fail (entrye->cboxtext != NULL);

  childs = gtk_combo_box_text_get_number_of_childs 
    (GTK_COMBO_BOX_TEXT (entrye->cboxtext));

  if (entrye->keep_history && entrye->history_length >= childs)
    gtk_combo_box_text_add_item (GTK_COMBO_BOX_TEXT (entrye->cboxtext),
				 string);
}

static gchar *
gtk_entry_extra_intersect (gint max_len, GSList *strings)
{
  gint   l = 0;
  gchar *completion;
  
  g_return_val_if_fail (max_len != 0, NULL);
  g_return_val_if_fail (strings != NULL, NULL);
  
  completion = g_new0 (gchar, max_len + 1);
  
  while (l < max_len)
    {
      gchar   *s = strings->data;
      GSList  *slist = NULL;
      
      s += l;
      completion[l] = *s;
      
      for (slist = strings->next; slist != NULL; slist = slist->next)
        {
          s = slist->data;
	  s += l;
	  
          if (completion[l] != *s)
            completion[l] = 0;
        }
      
      if (!completion[l])
        break;
      
      l++;
    }
  
  completion[l] = 0;
  
  return g_renew (gchar, completion, completion[0] ? l + 1 : 0);
}

/* private function */
void
gtk_entry_extra_register_combo_box_text (GtkEntryExtra *entrye,
					 GtkComboBoxText *cboxtext)
{
  g_return_if_fail (entrye != NULL);
  g_return_if_fail (GTK_IS_ENTRY_EXTRA (entrye));
  g_return_if_fail (cboxtext != NULL);
  g_return_if_fail (GTK_IS_COMBO_BOX_TEXT (cboxtext));

  gtk_entry_set_text (GTK_ENTRY (entrye), "");
  entrye->cboxtext = GTK_WIDGET (cboxtext);

  gtk_signal_connect (GTK_OBJECT (entrye),
		      "key_press_event",
		      GTK_SIGNAL_FUNC (gtk_entry_extra_key_press),
		      NULL);
}

static gint
gtk_entry_extra_key_press (GtkWidget       *widget,
			   GdkEventKey     *event)
{
  gboolean handled = FALSE;

  GtkEntryExtra *entry = GTK_ENTRY_EXTRA (widget);

  if (!entry->autocomplete)
    return handled;
  
  if ((event->keyval == GDK_Tab || event->keyval == GDK_ISO_Left_Tab)
      && !(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK
                           | GDK_MOD1_MASK)))
    {
      handled = TRUE;
      
      if (event->type == GDK_KEY_PRESS)
        {
          gchar *mstring;
          gchar *estring;
          
	  mstring = gtk_entry_extra_try_complete (entry);
	  estring = gtk_entry_get_text (GTK_ENTRY (entry));
          
          if (!gtk_entry_extra_match (entry,
				      estring))
            if (mstring && strcmp (mstring,
                                   estring)) {
              gtk_entry_set_text 
                (GTK_ENTRY (entry), mstring);
	      gtk_entry_set_position (GTK_ENTRY (entry),
				      GTK_ENTRY (entry)->text_length);
	    }
          
          g_free (mstring);
          
          gtk_signal_emit_stop_by_name 
            (GTK_OBJECT (entry), "key-press-event");
        }
    } 
  else if ((event->keyval == GDK_Return) 
           || (event->keyval == GDK_KP_Enter)) 
    {
      gchar *tmp;
      gchar *estring = gtk_entry_get_text (GTK_ENTRY (entry));
      
      handled = TRUE;
      
      if (estring && strlen (estring) > 1 
          && event->type == GDK_KEY_PRESS)
        {
          if (!entry->pattern_matching)
            {
              if (gtk_entry_extra_match (entry,
					 estring))
                {
                  tmp = gtk_entry_get_text (GTK_ENTRY (entry));
                  
                  gtk_signal_emit 
                    (GTK_OBJECT(entry),
                     gtk_entry_extra_signals[AUTOCOMPLETED], tmp);
                } 
              else
                {
                  tmp = gtk_entry_get_text (GTK_ENTRY (entry));
                  
                  if (entry->keep_history)
                    gtk_entry_extra_add_history (entry, tmp);
                }
            } 
          else
            gtk_combo_box_text_select_matches
	      (GTK_COMBO_BOX_TEXT (entry->cboxtext), estring);
        }
      
      gtk_signal_emit_stop_by_name (GTK_OBJECT (entry),
                                    "key-press-event");
    }
  
  return handled;
}


GtkType
gtk_entry_extra_get_type (void)
{
  static GtkType type = 0;
  
  if (!type) {
    static const GTypeInfo info = {
      sizeof (GtkEntryExtraClass),
      NULL, /* GBaseInitFunc */
      NULL, /* GBaseFinalizeFunc */
      (GClassInitFunc) gtk_entry_extra_class_init,
      NULL,
      NULL, /* class data */
      sizeof (GtkEntryExtra),
      0,
      (GInstanceInitFunc) gtk_entry_extra_init
    };

    type = g_type_register_static (gtk_entry_get_type (),
				   "GtkEntryExtra",
				   &info, 0);
  }

  return type;
}

GtkWidget *
gtk_entry_extra_new (void)
{
  GtkEntryExtra *entrye;
  
  entrye = gtk_type_new (gtk_entry_extra_get_type ());
  
  return GTK_WIDGET (entrye);
}

void
gtk_entry_extra_set_keep_history (GtkEntryExtra *entrye,
				  gboolean bool)
{
  g_return_if_fail (entrye != NULL);
  g_return_if_fail (GTK_IS_ENTRY_EXTRA (entrye));
  
  entrye->keep_history = bool;
}

void
gtk_entry_extra_set_history_length (GtkEntryExtra *entrye,
				    gint length)
{
  g_return_if_fail (entrye != NULL);
  g_return_if_fail (GTK_IS_ENTRY_EXTRA (entrye));
  g_return_if_fail (length > -1);
  
  /* FIXME: should we remove some items if
   * entrye->history_length > length ?
   */
  entrye->history_length = length;
}

void
gtk_entry_extra_set_autocomplete (GtkEntryExtra *entrye,
				  gboolean bool)
{
  g_return_if_fail (entrye != NULL);
  g_return_if_fail (GTK_IS_ENTRY_EXTRA (entrye));
  
  entrye->autocomplete = bool;
}

void
gtk_entry_extra_set_pattern_matching (GtkEntryExtra *entrye,
				      gboolean bool)
{
  g_return_if_fail (entrye != NULL);
  g_return_if_fail (GTK_IS_ENTRY_EXTRA (entrye));
  
  entrye->pattern_matching = bool;
}
